'use strict';

const _ = require('lodash');
const png = require('./png-image');
const buffer = require('./png-buffer');
const DiffArea = require('./diff-area');
const DiffClusters = require('./diff-clusters');
const validators = require('./validators');

exports.readPngCb = ({source, ...opts}) => {
    const readFunc = Buffer.isBuffer(source) ? png.fromBuffer : png.fromFile;
    return readFunc(source, opts);
};

exports.readBufferCb = ({source, ...opts}) => {
    const readFunc = Buffer.isBuffer(source) ? buffer.create : buffer.fromFile;
    return readFunc(source, opts);
};

exports.readPair = async (first, second, readCb = exports.readPngCb) => {
    const [firstPng, secondPng] = await Promise.all([first, second].map(readCb));

    return {first: firstPng, second: secondPng};
};

const getDiffClusters = (diffClusters, diffArea, {shouldCluster}) => {
    return shouldCluster ? diffClusters.clusters : [diffArea.area];
};

exports.getDiffPixelsCoords = async (png1, png2, predicate, opts = {}) => {
    const stopOnFirstFail = opts.hasOwnProperty('stopOnFirstFail') ? opts.stopOnFirstFail : false;

    const width = Math.min(png1.width, png2.width);
    const height = Math.min(png1.height, png2.height);

    const diffArea = new DiffArea();
    const diffClusters = new DiffClusters(opts.clustersSize);

    return new Promise((resolve) => {
        const processRow = (y) => {
            setImmediate(() => {
                for (let x = 0; x < width; x++) {
                    const color1 = png1.getPixel(x, y);
                    const color2 = png2.getPixel(x, y);

                    const result = predicate({
                        color1, color2,
                        png1, png2,
                        x, y,
                        width, height
                    });

                    if (!result) {
                        const {x: actX, y: actY} = png1.getActualCoord(x, y);
                        diffArea.update(actX, actY);
                        if (opts.shouldCluster) {
                            diffClusters.update(actX, actY);
                        }

                        if (stopOnFirstFail) {
                            return resolve({diffArea, diffClusters: getDiffClusters(diffClusters, diffArea, opts)});
                        }
                    }
                }

                y++;

                if (y < height) {
                    processRow(y);
                } else {
                    resolve({diffArea, diffClusters: getDiffClusters(diffClusters, diffArea, opts)});
                }
            });
        };

        processRow(0);
    });
};

exports.formatImages = (img1, img2) => {
    validators.validateImages(img1, img2);

    return [img1, img2].map((i) => {
        return _.isObject(i) && !Buffer.isBuffer(i) ? i : {source: i, boundingBox: null};
    });
};

exports.areBuffersEqual = (img1, img2) => {
    if (img1.boundingBox || img2.boundingBox) {
        return false;
    }

    return img1.buffer.equals(img2.buffer);
};
